"""
The main entrypoint for the prefect flow
"""

from typing import Any, Literal

from prefect import flow, get_run_logger, task
from prefect.task_runners import ConcurrentTaskRunner

from investigraph import __version__, settings
from investigraph.logic.aggregate import in_memory
from investigraph.logic.extract import iter_records
from investigraph.logic.fetch import fetch_source, get_cache_key
from investigraph.model import (
    Context,
    Flow,
    FlowOptions,
    HttpSourceResponse,
    SmartSourceResponse,
)
from investigraph.model.context import init_context
from investigraph.util import get_func


@task
def aggregate(ctx: Context):
    logger = get_run_logger()
    fragments, proxies = in_memory(ctx, ctx.config.fragments_uri)
    logger.info("AGGREGATED %d fragments to %d proxies", fragments, proxies)
    out = ctx.config.entities_uri
    logger.info("OUTPUT: %s", out)
    return out


@task
def load(ctx: Context, ckey: str):
    logger = get_run_logger()
    proxies = ctx.cache.get(ckey)
    out = ctx.config.fragments_uri
    if ctx.config.target == "postgres":
        # write directly to entities instead of fragments
        # as aggregation is happening within postgres store on write
        out = ctx.entities_loader.write(proxies)
    else:
        out = ctx.fragments_loader.write(proxies)
    logger.info("LOADED %d proxies", len(proxies))
    logger.info("OUTPUT: %s", out)
    return out


@task
def transform(ctx: Context, ckey: str) -> str:
    logger = get_run_logger()
    parse_record = get_func(ctx.config.parse_module_path)
    proxies: list[dict[str, Any]] = []
    records = ctx.cache.get(ckey)
    for rec in records:
        for proxy in parse_record(ctx, rec):
            proxies.append(proxy.to_dict())
    logger.info("TRANSFORMED %d records", len(records))
    return ctx.cache.set(proxies)


@task(
    retries=settings.FETCH_RETRIES,
    retry_delay_seconds=settings.FETCH_RETRY_DELAY,
    cache_key_fn=get_cache_key,
)
def fetch(ctx: Context) -> Literal[HttpSourceResponse, SmartSourceResponse]:
    logger = get_run_logger()
    logger.info("OPEN %s", ctx.source.uri)
    return fetch_source(ctx.source)


@flow(
    name="investigraph-pipeline",
    version=__version__,
    flow_run_name="{ctx.dataset}-{ctx.source.name}",
    task_runner=ConcurrentTaskRunner(),
)
def run_pipeline(ctx: Context):
    res = fetch.submit(ctx)
    res = res.result()
    ix = 0
    batch = []
    results = []
    for ix, rec in enumerate(iter_records(res)):
        batch.append(rec)
        if ix and ix % 1000 == 0:
            results.append(transform.submit(ctx, ctx.cache.set(batch)))
            batch = []
    if batch:
        results.append(transform.submit(ctx, ctx.cache.set(batch)))

    logger = get_run_logger()
    logger.info("EXTRACTED %d records", ix + 1)

    # write
    for res in results:
        res = res.result()
        load.submit(ctx, res)


@flow(
    name="investigraph",
    version=__version__,
    flow_run_name="{options.dataset}",
)
def run(options: FlowOptions):
    flow = Flow.from_options(options)
    for source in flow.config.pipeline.sources:
        ctx = init_context(config=flow.config, source=source)
        ctx.export_metadata()
        run_pipeline(ctx)

    if flow.should_aggregate:
        aggregate(ctx)



"""
Transform stage: map data records to ftm proxies
"""

from followthemoney.mapping import QueryMapping

from investigraph.model import Context
from investigraph.types import CEGenerator, SDict
from investigraph.util import uplevel_proxy


def map_record(record: SDict, mapping: QueryMapping) -> CEGenerator:
    if mapping.source.check_filters(record):
        entities = mapping.map(record)
        for proxy in entities.values():
            proxy = uplevel_proxy(proxy)
            yield proxy


def map_ftm(ctx: Context, data: SDict) -> CEGenerator:
    for mapping in ctx.config.mappings:
        yield from map_record(data, mapping)



from datetime import datetime
from typing import Any, Dict, Generator, Optional, Union

import requests
import yaml
from nomenklatura.dataset import DataCatalog
from nomenklatura.util import PathLike, datetime_iso
from zavod.dataset import ZavodDataset

from .logging import get_logger

log = get_logger(__name__)


def flatten_catalog(url: str) -> Generator[Dict[str, Any], None, None]:
    try:
        resp = requests.get(url)
        catalog_data = resp.json()
        for dataset in catalog_data.get("datasets", []):
            if dataset.get("type") != "collection":
                yield dataset
    except Exception as exc:
        log.error(str(exc), url=url)


def build_catalog(catalog_in: PathLike) -> DataCatalog:
    """
    build a catalog based on yaml input:

    ```yaml
    datasets:
      - include: https://data.ftm.store/investigraph/ec_meetings/index.json
    - - include_catalog: https://data.ftm.store/catalog.json
      - include_catalog:
          url: https://data.opensanctions.org/datasets/latest/index.json
          exclude:
            - validation
          maintainer:
            name: OpenSanctions
            url: https://opensanctions.org
            description: |
              OpenSanctions is an international database of persons and companies
              of political, criminal, or economic interest. The project combines
              the sanctions lists, databases of politically exposed persons, and
              other information about persons in the public interest into a single,
              easy-to-use dataset.

    ```

    """
    log.info("building catalog", catalog=str(catalog_in))
    seen = set()
    with open(catalog_in, "r") as fh:
        catalog_in_data = yaml.safe_load(fh)
    catalog_in = str(catalog_in)  # for logging
    catalog = DataCatalog(ZavodDataset, {})
    catalog.updated_at = datetime_iso(datetime.utcnow())
    for ds_data in catalog_in_data["datasets"]:
        include_url: Optional[str] = ds_data.pop("include", None)
        if include_url is not None:
            try:
                resp = requests.get(include_url)
                ds_data = resp.json()
                name = ds_data["name"]
                if name not in seen:
                    ds = catalog.make_dataset(ds_data)
                    log.info("Dataset: %r" % ds, catalog=catalog_in)
                    seen.add(name)
                else:
                    log.warn(
                        "Dataset `%s` already in catalog." % name, catalog=catalog_in
                    )
            except Exception as exc:
                log.error(str(exc), catalog=catalog_in, url=include_url)
        include_catalog: Optional[Union[str, Dict[str, Any]]] = ds_data.pop(
            "include_catalog", None
        )
        if include_catalog is not None:
            if isinstance(include_catalog, str):
                include_catalog_url = include_catalog
                exclude = []
            else:
                include_catalog_url = include_catalog.pop("url")
                exclude = include_catalog.pop("exclude", [])

            for ds_data in flatten_catalog(include_catalog_url):
                name = ds_data["name"]
                if name not in seen and name not in exclude:
                    ds = catalog.make_dataset(ds_data)
                    log.info("Dataset: %r" % ds, catalog=catalog_in)
                    seen.add(name)
                else:
                    log.warn(
                        "Dataset `%s` already in catalog." % name, catalog=catalog_in
                    )

    return catalog



import mimetypes
from datetime import datetime

import requests
from dateparser import parse as parse_date
from normality import slugify
from pantomime import normalize_mimetype, types
from pydantic import BaseModel
from smart_open import open, parse_uri

from investigraph.exceptions import ImproperlyConfigured
from investigraph.types import BytesGenerator, SDict
from investigraph.util import slugified_dict


class SourceHead(BaseModel):
    etag: str | None = None
    last_modified: datetime | None = None
    content_type: str | None = None
    content_length: int | None = None

    def __init__(self, **data):
        super().__init__(
            last_modified=parse_date(data.pop("last_modified", "")),
            content_type=normalize_mimetype(data.pop("content_type", None)),
            **data,
        )

    def should_stream(self) -> bool:
        return self.content_type in (types.CSV, types.JSON)


class Source(BaseModel):
    name: str
    uri: str
    scheme: str
    extract_kwargs: dict | None = {}

    def __init__(self, **data):
        data["uri"] = str(data["uri"])
        data["name"] = data.get("name", slugify(data["uri"]))
        data["scheme"] = data.get("scheme", parse_uri(data["uri"]).scheme)
        super().__init__(**data)

    @property
    def is_http(self) -> bool:
        return self.scheme.startswith("http")

    def head(self) -> SourceHead:
        if self.is_http:
            res = requests.head(self.uri)
            return SourceHead(**slugified_dict(res.headers))
        raise NotImplementedError("Cannot fetch head for scheme %s" % self.scheme)

    def iter_lines(self) -> BytesGenerator:
        raise NotImplementedError


class HttpSourceResponse(Source):
    is_stream: bool | None = False
    response: requests.Response
    header: SDict

    class Config:
        arbitrary_types_allowed = True

    @property
    def mimetype(self) -> str:
        return normalize_mimetype(self.header["content_type"])

    @property
    def content(self) -> bytes:
        if self.is_stream:
            raise ImproperlyConfigured("%s is a stream" % self.uri)
        return self.response.content

    def iter_lines(self) -> BytesGenerator:
        yield from self.response.iter_lines()


class SmartSourceResponse(Source):
    @property
    def content(self) -> bytes:
        if self.is_stream:
            raise ImproperlyConfigured("%s is a stream" % self.uri)
        with open(self.uri, "rb") as fh:
            return fh.read()

    def iter_lines(self) -> BytesGenerator:
        with open(self.uri, "rb") as fh:
            for line in fh:
                yield line

    @property
    def mimetype(self) -> str:
        mtype, _ = mimetypes.guess_type(self.uri)
        return normalize_mimetype(mtype)

    @property
    def is_stream(self) -> bool:
        return self.mimetype in (types.CSV, types.JSON)



from banal import clean_dict
from pydantic import BaseModel

from .config import Config, get_config


class FlowOptions(BaseModel):
    dataset: str
    block: str | None = None
    config: str | None = None
    index_uri: str | None = None
    fragments_uri: str | None = None
    entities_uri: str | None = None
    aggregate: bool | None = True


class Flow(BaseModel):
    dataset: str
    config: Config

    def __init__(self, **data):
        # override base config with runtime options
        options = data.get("options")
        if options is not None:
            options = dict(options)
        config = get_config(
            data["dataset"], options.get("block"), options.get("config")
        )
        data["config"] = {**clean_dict(config.dict()), **clean_dict(options)}
        super().__init__(**data)

    @property
    def should_aggregate(self) -> bool:
        if self.config.entities_uri.startswith("postg"):
            return False
        return self.config.aggregate

    @classmethod
    def from_options(cls, options: FlowOptions) -> "Flow":
        return cls(dataset=options.dataset, options=options)



import logging
import sys
from typing import List

import structlog
from structlog.contextvars import merge_contextvars
from structlog.stdlib import get_logger as get_raw_logger
from structlog.types import Processor


def configure_logging(
    level: int = logging.INFO,
    extra_processors: List[Processor] = [],
) -> None:
    """Configure log levels and structured logging."""
    processors: List[Processor] = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        merge_contextvars,
    ]
    processors.extend(extra_processors)
    renderer = structlog.dev.ConsoleRenderer(
        exception_formatter=structlog.dev.plain_traceback
    )
    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=list(processors),
        processor=renderer,
    )

    processors.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)

    # configuration for structlog based loggers
    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(level)
    handler.setFormatter(formatter)

    logger = logging.getLogger("investigraph")
    logger.setLevel(level)
    logger.handlers = [handler]


def get_logger(name: str) -> structlog.stdlib.BoundLogger:
    return get_raw_logger(name)



from functools import cache
from pathlib import Path
from typing import Any

import yaml
from banal import keys_values
from followthemoney import model
from followthemoney.mapping import QueryMapping
from nomenklatura.dataset.catalog import DataCatalog
from nomenklatura.dataset.dataset import Dataset
from nomenklatura.util import PathLike
from pydantic import BaseModel
from smart_open import open

from investigraph.exceptions import ImproperlyConfigured
from investigraph.logging import get_logger
from investigraph.settings import DATASETS_BLOCK, DEFAULT_TRANSFORMER
from investigraph.util import ensure_pythonpath

from .block import get_block
from .source import Source

log = get_logger(__name__)


class Pipeline(BaseModel):
    sources: list[Source]


class Config(BaseModel):
    dataset: str
    base_path: Path
    metadata: dict[str, Any]
    mappings: list[QueryMapping]
    pipeline: Pipeline
    index_uri: str | None = None
    fragments_uri: str | None = None
    entities_uri: str | None = None
    aggregate: bool | None = True

    class Config:
        arbitrary_types_allowed = True

    @property
    def target(self) -> str:
        if self.entities_uri is not None:
            if self.entities_uri.startswith("postg"):
                return "postgres"
        return "json"

    @property
    def parse_module_path(self) -> str:
        if len(self.mappings):
            return DEFAULT_TRANSFORMER
        return f"{self.base_path.parent.name}.{self.dataset}.parse:parse"

    @classmethod
    def from_path(cls, fp: PathLike) -> "Config":
        base_path = Path(fp).parent
        ensure_pythonpath(base_path.parent.parent)
        catalog = DataCatalog(Dataset, {})
        with open(fp, "r") as fh:
            data = yaml.safe_load(fh)
        data["title"] = data.get("title", data["name"].title())
        dataset: Dataset = catalog.make_dataset(data)
        mappings = []
        if data.get("mapping") is not None:
            for m in keys_values(data["mapping"], "queries", "query"):
                m.pop("database", None)
                m["csv_url"] = "/dev/null"
                mapping = model.make_mapping(m)
                mappings.append(mapping)

        config = {
            "dataset": dataset.name,
            "base_path": base_path,
            "metadata": dataset.to_dict(),
            "mappings": mappings,
        }

        for key in cls.__fields__:
            if key not in config:
                config[key] = data.get(key)

        return cls(**config)


@cache
def get_config(
    dataset: str | None = None, block: str | None = None, path: PathLike | None = None
) -> Config:
    """
    Return configuration based on block or path (path has precedence)
    """
    if path is not None:
        return Config.from_path(path)
    if dataset is not None:
        if block is not None:
            block = get_block(block)
        else:
            block = get_block(DATASETS_BLOCK)
        log.info("Using block `%s`" % block)
        block.load(dataset)
        return Config.from_path(block.path / dataset / "config.yml")
    raise ImproperlyConfigured("Specify `dataset` and `block` or `path` to config.")



"""
Load transformed data (aka CE proxies) to various targets like json files,
databases, s3 endpoints...
"""

from functools import cache
from typing import Iterable, TypeAlias
from urllib.parse import urlparse

import shortuuid
from ftmstore import get_dataset

from investigraph.cache import get_cache
from investigraph.types import SDict
from investigraph.util import smart_write_proxies

Proxies: TypeAlias = Iterable[SDict]


def to_uri(uri: str, proxies: Proxies, **kwargs):
    smart_write_proxies(uri, proxies, **kwargs)


def to_store(uri: str, dataset: str, proxies: Proxies):
    dataset = get_dataset(dataset, database_uri=uri)
    bulk = dataset.bulk()
    for ix, proxy in enumerate(proxies):
        bulk.put(proxy, fragment=str(ix))
    bulk.flush()


class Loader:
    """
    write entity proxies to anywhere
    """

    def __init__(self, uri: str, dataset: str, parts: bool | None = False):
        self.uri = uri
        self.dataset = dataset
        parsed = urlparse(uri)
        self.is_store = "sql" in parsed.scheme
        self.cache = get_cache()
        self.parts = parts

    def write(self, proxies: Proxies, **kwargs) -> None:
        if self.is_store:
            to_store(self.uri, self.dataset, proxies)
            return self.uri
        else:
            uri = self.uri
            if self.parts:
                uri += f".{shortuuid.uuid()}"
                self.cache.sadd(uri, key=self.uri)
            to_uri(uri, proxies, **kwargs)
            return uri


@cache
def get_loader(uri: str, dataset: str, parts: bool | None = False) -> Loader:
    return Loader(uri, dataset, parts)



from .block import DatasetBlock  # noqa
from .config import Config  # noqa
from .context import Context  # noqa
from .flow import Flow, FlowOptions  # noqa
from .source import HttpSourceResponse, SmartSourceResponse, Source, SourceHead  # noqa



from typing import Any, Generator, TypeAlias

from nomenklatura.entity import CE

# a string-keyed dict
SDict: TypeAlias = dict[str, Any]
RecordGenerator: TypeAlias = Generator[SDict, None, None]

# composite entity generator
CEGenerator: TypeAlias = Generator[CE, None, None]

StrGenerator: TypeAlias = Generator[str, None, None]
BytesGenerator: TypeAlias = Generator[bytes, None, None]



from functools import cache
from typing import Any, Iterable, Set

import fakeredis
import redis
import shortuuid
from cachelib.serializers import RedisSerializer

from investigraph import settings
from investigraph.logging import get_logger

log = get_logger(__name__)


class Cache:
    """
    This is an extremly simple cache interface for sharing tasks data
    efficiently via redis (or fakeredis during development)

    it creates (prefixed) random keys during data set to cache.

    it mimics redis GETDEL so that after fetching data from cache the key is
    deleted (turn of by `delete=False`)
    """

    serializer = RedisSerializer()

    def __init__(self):
        if settings.DEBUG:
            con = fakeredis.FakeStrictRedis()
        else:
            con = redis.from_url(settings.REDIS_URL)
        con.ping()
        log.info("Redis initialized", url=settings.REDIS_URL)
        self.cache = con

    def set(self, data: Any, key: str | None = None) -> str:
        data = self.serializer.dumps(data)
        key = key or shortuuid.uuid()
        self.cache.set(self.get_key(key), data)
        return key

    def get(self, key: str, delete: bool | None = True) -> Any:
        key = self.get_key(key)
        res = self.cache.get(key)
        if delete:
            self.cache.delete(key)  # GETDEL
        if res is not None:
            data = self.serializer.loads(res)
            return data

    def sadd(self, *values: Iterable[Any], key: str | None = None) -> str:
        key = key or shortuuid.uuid()
        values = [str(v) for v in values]
        self.cache.sadd(self.get_key(key) + "#SET", *values)
        return key

    def smembers(self, key: str, delete: bool | None = True) -> Set[str]:
        key = self.get_key(key) + "#SET"
        res: Set[bytes] = self.cache.smembers(key)
        if delete:
            self.cache.delete(key)
        return {v.decode() for v in res} or None

    def flushall(self):
        self.cache.flushall()

    @staticmethod
    def get_key(key: str) -> str:
        return f"{settings.CACHE_PREFIX}:{key}"


@cache
def get_cache() -> Cache:
    return Cache()



from datetime import datetime

import orjson
import shortuuid
from nomenklatura.util import datetime_iso
from prefect.runtime import flow_run
from pydantic import BaseModel
from smart_open import open

from investigraph.cache import Cache, get_cache
from investigraph.logic.load import Loader, get_loader
from investigraph.settings import DATA_ROOT
from investigraph.util import ensure_path

from .config import Config
from .source import Source


class Context(BaseModel):
    dataset: str
    prefix: str
    config: Config
    source: Source
    run_id: str
    entities_loader: Loader
    fragments_loader: Loader

    class Config:
        arbitrary_types_allowed = True

    @property
    def cache(self) -> Cache:
        return get_cache()

    def export_metadata(self) -> None:
        data = self.config.metadata
        data["updated_at"] = data.get("updated_at", datetime_iso(datetime.utcnow()))
        data = orjson.dumps(data)
        with open(self.config.index_uri, "wb") as fh:
            fh.write(data)


def init_context(config: Config, source: Source) -> Context:
    run_id = flow_run.get_id() or f"dummy-{shortuuid.uuid()}"
    path = ensure_path(DATA_ROOT / config.dataset)
    if config.index_uri is None:
        config.index_uri = (path / "index.json").as_uri()
    if config.fragments_uri is None:
        config.fragments_uri = (path / f"fragments.{run_id}.json").as_uri()
    if config.entities_uri is None:
        config.entities_uri = (path / "entities.ftm.json").as_uri()

    return Context(
        dataset=config.dataset,
        prefix=config.metadata.get("prefix", config.dataset),
        config=config,
        source=source,
        run_id=run_id,
        fragments_loader=get_loader(config.fragments_uri, config.dataset, parts=True),
        entities_loader=get_loader(config.entities_uri, config.dataset),
    )



"""
Extract sources to iterate objects to dict records
"""

from io import BytesIO, StringIO

import pandas as pd
from pantomime import types

from investigraph.model import Source
from investigraph.types import RecordGenerator

TABULAR = [types.CSV, types.EXCEL, types.XLS, types.XLSX]


def read_pandas(mimetype: str, content: str | bytes, **kwargs) -> pd.DataFrame:
    if isinstance(content, bytes):
        content = BytesIO(content)
    else:
        content = StringIO(content)

    if mimetype in (types.XLS, types.XLSX, types.EXCEL):
        return pd.read_excel(content, **kwargs).fillna("")
    return pd.read_csv(content, **kwargs).fillna("")


def yield_pandas(df: pd.DataFrame) -> RecordGenerator:
    for _, row in df.iterrows():
        yield dict(row)


def iter_records(res: Source) -> RecordGenerator:
    if res.mimetype in TABULAR:
        kwargs = {**{"dtype": str}, **res.extract_kwargs}
        if res.is_stream:
            # yield pandas chunks to be able to apply extract_kwargs
            # doesn't work for excel here
            lines = []
            for ix, line in enumerate(res.iter_lines()):
                lines.append(line)
                if ix and ix % 10_000 == 0:
                    content = b"\r".join(lines)
                    df = read_pandas(res.mimetype, content, **kwargs)
                    lines = []
                    if ix == 10_000:  # first chunk
                        # fix initial kwargs for next chunk
                        kwargs["names"] = df.columns
                        kwargs["header"] = 0
                        kwargs.pop("skiprows")  # FIXME what else?
                    yield from yield_pandas(df)
            if lines:
                content = b"\r".join(lines)
                df = read_pandas(res.mimetype, content, **kwargs)
                yield from yield_pandas(df)
        else:
            df = read_pandas(res.mimetype, res.content, **kwargs)
            yield from yield_pandas(df)
        return

    raise NotImplementedError("unsupported mimetype: `%s`" % res.mimetype)



"""
aggregate fragments
"""

import logging
from uuid import uuid4

from ftmstore import get_dataset

from investigraph.cache import get_cache
from investigraph.model import Context
from investigraph.types import CEGenerator
from investigraph.util import smart_iter_proxies

log = logging.getLogger(__name__)


def get_smart_proxies(uri: str) -> CEGenerator:
    """
    see if we have parts in cache during run time
    (mimics efficient globbing for remote sources)
    """
    cache = get_cache()
    uris = cache.smembers(uri)
    if uris:
        for uri in uris:
            yield from smart_iter_proxies(uri)
        return

    yield from smart_iter_proxies(uri)


def in_memory(ctx: Context, in_uri: str) -> tuple[int, int]:
    """
    aggregate in memory: read fragments from `in_uri` and write aggregated
    proxies to `out_uri`

    as `smart_open` is used here, `in_uri` and `out_uri` can be any local or
    remote locations
    """
    fragments = 0
    buffer = {}
    for proxy in get_smart_proxies(in_uri):
        fragments += 1
        if proxy.id in buffer:
            buffer[proxy.id].merge(proxy)
        else:
            buffer[proxy.id] = proxy

    ctx.entities_loader.write(buffer.values(), serialize=True)
    return fragments, len(buffer.values())


def in_db(ctx: Context, in_uri: str) -> tuple[int, int]:
    """
    use ftm store database to aggregate.
    `in_uri`: database connection string
    """
    dataset = get_dataset("aggregate_%s" % uuid4().hex)
    bulk = dataset.bulk()
    for ix, proxy in enumerate(get_smart_proxies(in_uri)):
        if ix % 10000 == 0:
            log.info("Write [%s]: %s entities", dataset.name, ix)
        bulk.put(proxy, fragment=str(ix))
    bulk.flush()
    proxies = []
    for ox, proxy in enumerate(dataset.iterate()):
        proxies.append(proxy)
        if ox % 10000 == 0:
            ctx.entities_loader.write(proxies, serialize=True)
            proxies = []
    ctx.entities_loader.write(proxies, serialize=True)
    dataset.drop()
    return ix, ox



"""
Utils to fetch remote sources, optionally respect cache headers
"""

from datetime import datetime
from typing import Literal

import requests

from investigraph.model import Context, HttpSourceResponse, SmartSourceResponse, Source
from investigraph.util import slugified_dict


def fetch_source(source: Source) -> Literal[HttpSourceResponse, SmartSourceResponse]:
    if source.is_http:
        head = source.head()
        stream = head.should_stream()
        res = requests.get(source.uri, stream=stream)
        assert res.ok
        return HttpSourceResponse(
            **source.dict(),
            header=slugified_dict(res.headers),
            response=res,
            is_stream=stream,
        )
    return SmartSourceResponse(**source.dict())


def get_cache_key(_, params) -> str:
    """
    cache results of fetch tasks based on etag or last_modified
    """
    ctx: Context = params["ctx"]
    if ctx.source.is_http:
        head = ctx.source.head()
        url = ctx.source.uri
        if head.etag:
            return f"{url}-{head.etag}"
        if head.last_modified:
            return f"{url}-{head.last_modified}"
    return f"{url}-{datetime.now()}"  # actually don't cache



import sys
from functools import cache
from importlib import import_module, invalidate_caches
from pathlib import Path
from typing import Any, Callable, Iterable

import orjson
from banal import ensure_dict
from followthemoney import model
from followthemoney.proxy import E
from nomenklatura.entity import CE, CompositeEntity
from normality import slugify
from smart_open import open

from investigraph.types import CEGenerator, SDict


def slugified_dict(data: dict[Any, Any]) -> SDict:
    return {slugify(k, "_"): v for k, v in ensure_dict(data).items()}


def make_proxy(schema: str) -> CE:
    return CompositeEntity.from_dict(model, {"schema": schema})


def uplevel_proxy(proxy: E) -> CE:
    return CompositeEntity.from_dict(model, proxy.to_dict())


def load_proxy(data: SDict) -> CE:
    return CompositeEntity.from_dict(model, data)


def ensure_proxy(data: SDict | CE) -> CE:
    if isinstance(data, CE):
        return data
    return load_proxy(data)


def smart_iter_proxies(uri: str) -> CEGenerator:
    for line in open(uri):
        yield load_proxy(orjson.loads(line))


def smart_write_proxies(
    uri: str,
    proxies: Iterable[CE | SDict],
    mode: str | None = "wb",
    serialize: bool | None = False,
) -> int:
    with open(uri, mode) as f:
        for ix, proxy in enumerate(proxies):
            if serialize:
                proxy = proxy.to_dict()
            f.write(orjson.dumps(proxy, option=orjson.OPT_APPEND_NEWLINE))
    return ix + 1


@cache
def ensure_path(path: Path) -> Path:
    path.mkdir(parents=True, exist_ok=True)
    return path.absolute()


@cache
def ensure_pythonpath(path: Path) -> None:
    path = str(path)
    if path not in sys.path:
        sys.path.append(path)


@cache
def get_func(parse_module_path: str) -> Callable:
    invalidate_caches()
    module, func = parse_module_path.split(":")
    module = import_module(module)
    return getattr(module, func)



"""
Inspect dataset pipelines interactively
"""


from typing import Any, Generator

import pandas as pd
from nomenklatura.entity import CE
from nomenklatura.util import PathLike
from rich import print

from investigraph.logic.extract import iter_records
from investigraph.logic.fetch import fetch_source
from investigraph.model.config import Config, get_config
from investigraph.model.context import init_context
from investigraph.model.source import Source
from investigraph.util import get_func


def print_error(msg: str):
    print(f"[bold red]ERROR[/bold red] {msg}")


def get_records(source: Source) -> list[dict[str, Any]]:
    records: list[dict[str, Any]] = []
    print("Fetching `%s` ..." % source.uri)
    res = fetch_source(source)
    for ix, rec in enumerate(iter_records(res)):
        records.append(rec)
        if ix == 5:
            return records


def inspect_config(p: PathLike) -> Config:
    config = get_config(path=p)
    try:
        func = get_func(config.parse_module_path)
        if not callable(func):
            print_error(
                f"module not found or not callable: `{config.parse_module_path}`"
            )
    except ModuleNotFoundError:
        print_error(f"no parsing function: `{config.parse_module_path}`")
    return config


def inspect_extract(config: Config) -> Generator[tuple[str, pd.DataFrame], None, None]:
    """
    Preview fetched & extracted records in tabular format
    """
    for source in config.pipeline.sources:
        df = pd.DataFrame(get_records(source))
        yield source.name, df


def inspect_transform(config: Config) -> Generator[tuple[str, CE], None, None]:
    """
    Preview first proxies
    """
    func = get_func(config.parse_module_path)
    for source in config.pipeline.sources:
        ctx = init_context(config, source)
        proxies: list[CE] = []
        for rec in get_records(source):
            for proxy in func(ctx, rec):
                proxies.append(proxy)
        yield source.name, proxies



import os
from pathlib import Path
from typing import Any

from banal import as_bool

from investigraph import __version__


def get_env(env: str, default: Any | None = None) -> Any | None:
    return os.environ.get(env, default)


DEBUG = as_bool(get_env("DEBUG", 1))
DATA_ROOT = Path(get_env("DATA_ROOT", Path.cwd() / "data")).absolute()
DATASETS_REPO = get_env(
    "DATASETS_REPO", "https://github.com/investigativedata/investigraph-datasets.git"
)
DATASETS_BLOCK = get_env("DATASETS_BLOCK", "github/investigraph-datasets")

DEFAULT_TRANSFORMER = get_env(
    "DEFAULT_TRANSFORMER", "investigraph.logic.transform:map_ftm"
)

REDIS_URL = get_env("REDIS_URL", "redis://localhost:6379")
CACHE_PREFIX = get_env("CACHE_PREFIX", f"investigraph:{__version__}")

FETCH_RETRIES = int(get_env("FETCH_RETRIES", 3))
FETCH_RETRY_DELAY = int(get_env("FETCH_RETRY_DELAY", 5))
